#include "EXTERN.h"
#include "perl.h"
#include "XSUB.h"

#undef Fflush
#undef Mkdir
#undef Stat

#include <stdio.h>
#include <string.h>
#include <utime.h>
#include "rpmlib.h"
#include <rpmio.h>
#include "rpmcli.h"

#include "rpmts.h"
#include "rpmte.h"

#include "rpmds.h"
#include "rpmfi.h"

#include "header.h"
#include "rpmdb.h"
#include "misc.h"

#include "rpmxs.h"

MODULE = RPM::Header		PACKAGE = RPM::Header

PROTOTYPES: ENABLE

void
stream2header(fp, callback = NULL)
    FILE *fp
    SV * callback
    PREINIT:
    FD_t fd = NULL;
    Header header;
    PPCODE:
    if (fp && (fd = fdDup(fileno(fp)))) {
        while ((header = headerRead(fd))) {
            if (callback != NULL && SvROK(callback)) {
                ENTER;
                SAVETMPS;
                PUSHMARK(SP);
                XPUSHs(sv_2mortal(sv_setref_pv(newSVpv("", 0), "RPM::Header", (void *)header)));
                PUTBACK;
                call_sv(callback, G_DISCARD | G_SCALAR);
                SPAGAIN;
                FREETMPS;
                LEAVE;
            } else {
                XPUSHs(sv_2mortal(sv_setref_pv(newSVpv("", 0), "RPM::Header", (void *)header)));
            }
        }
        Fclose(fd);
    }

# Read a rpm and return a Header
# Return undef if failed
void
rpm2header(filename, sv_vsflags = NULL)
    char * filename
    SV * sv_vsflags
    PREINIT:
    rpmts ts = rpmtsCreate();
    rpmVSFlags vsflags = RPMVSF_DEFAULT; 
    PPCODE:
    if (sv_vsflags == NULL) /* Nothing has been passed, default is no signature */
        vsflags |= _RPMVSF_NOSIGNATURES;
    else
        vsflags = sv2constant(sv_vsflags, "rpmvsflags");
    rpmtsSetVSFlags(ts, vsflags);
    _rpm2header(ts, filename, 0);
    SPAGAIN;
    ts = rpmtsFree(ts);

void
DESTROY(h)
	Header h
    CODE:
	headerFree(h);

#
# OO methodes
#

#
# Comparison functions
#

int
compare(h1, h2)
	Header h1
	Header h2
    CODE:
	RETVAL = rpmVersionCompare(h1, h2);
    OUTPUT:
    RETVAL


int
_op_spaceship(h1, h2, ...)
	Header h1
	Header h2
    CODE:
	RETVAL = rpmVersionCompare(h1, h2);
    OUTPUT:
    RETVAL

int
is_source_package(h)
	Header h
    CODE:
	RETVAL = !headerIsEntry(h, RPMTAG_SOURCERPM);
    OUTPUT:
	RETVAL

void
tagformat(h, format)
	Header h
	char * format
    PREINIT:
	const char * s;
    PPCODE:
	s =  headerSprintf(h, format, rpmTagTable, rpmHeaderFormats, NULL);
    if (s)
	    PUSHs(sv_2mortal(newSVpv((char *)s, 0)));
	s = _free(s); 

# Write rpm header into file pointer
# fedora use HEADER_MAGIC_NO, too bad, set no_header_magic make the function
# compatible
int
write(h, fp, no_header_magic = 0)
    Header h
    FILE * fp
    int no_header_magic
    PREINIT:
    FD_t fd;
    CODE:
    RETVAL = 0;
    if (h) {
        if ((fd = fdDup(fileno(fp))) != NULL) {
            headerWrite(fd, h);
            Fclose(fd);
            RETVAL = 1;
        }
    }
    OUTPUT:
    RETVAL

void
hsize(h, no_header_magic = 0)
    Header h
    int no_header_magic
    PPCODE:
    XPUSHs(sv_2mortal(newSViv(headerSizeof(h))));
    
void
copy(h)
    Header h
    PREINIT:
    Header hcopy;
    PPCODE:
    hcopy = headerCopy(h);
    XPUSHs(sv_2mortal(sv_setref_pv(newSVpv("", 0), "RPM::Header", (void *)hcopy)));
#ifdef HDRPMMEM
    PRINTF_NEW("RPM::Header", hcopy, hcopy->nrefs);
#endif

void
string(h, no_header_magic = 0)
    Header h
    int no_header_magic
    PREINIT:
    char * string = NULL;
    int offset = 8; /* header magic length */
    char * ptr = NULL;
    int hsize = 0;
    PPCODE:
    hsize = headerSizeof(h);
    string = headerUnload(h, NULL);
    if (! no_header_magic) {
        ptr = malloc(hsize);
        memcpy(ptr, header_magic, 8);
        memcpy(ptr + 8, string, hsize - 8);
    }
    XPUSHs(sv_2mortal(newSVpv(ptr ? ptr : string, hsize)));
    free(string);
    free(ptr);

int
removetag(h, sv_tag)
    Header h
    SV * sv_tag
    PREINIT:
    rpmTag tag = -1;
    CODE:
    if (SvIOK(sv_tag)) {
        tag = SvIV(sv_tag);
    } else if (SvPOK(sv_tag)) {
        tag = tagValue(SvPV_nolen(sv_tag));
    }
    if (tag > 0)
        RETVAL = headerRemoveEntry(h, tag);
    else
        RETVAL = 1;
    OUTPUT:
    RETVAL

int
addtag(h, sv_tag, sv_tagtype, ...)
    Header h
    SV * sv_tag
    SV * sv_tagtype
    PREINIT:
    HE_t he = memset(alloca(sizeof(*he)), 0, sizeof(*he));
    char * value;
    uint32_t ivalue;
    int i;
    STRLEN len;
    CODE:
    he->tag = -1;
    if (SvIOK(sv_tag)) {
        he->tag = SvIV(sv_tag);
    } else if (SvPOK(sv_tag)) {
        he->tag = tagValue(SvPV_nolen(sv_tag));
    }
    he->t = sv2constant(sv_tagtype, "rpmtagtype");
    if (he->tag > 0)
        RETVAL = 1;
    else
        RETVAL = 0;
    /* if (he->tag == RPMTAG_OLDFILENAMES)
        expandFilelist(h); */
    for (i = 3; (i < items) && RETVAL; i++) {
        switch (he->t) {
            case RPM_CHAR_TYPE:
            case RPM_UINT8_TYPE:
            case RPM_UINT16_TYPE:
            case RPM_UINT32_TYPE:
                ivalue = SvUV(ST(i));
		he->p.ui32p = &ivalue;
		he->c = 1;
		he->append = 1;
                RETVAL = headerAddOrAppendEntry(h, he->tag, he->t, he->p.ptr, he->c);
		he->append = 0;
                break;
            case RPM_BIN_TYPE:
                value = (char *)SvPV(ST(i), len);
		he->p.ptr = &value;
		he->c = len;
                RETVAL = headerAddEntry(h, he->tag, he->t, he->p.ptr, he->c);
                break;
            case RPM_STRING_ARRAY_TYPE:
                value = SvPV_nolen(ST(i));
		he->p.ptr = &value;
		he->c = 1;
		he->append = 1;
                RETVAL = headerAddOrAppendEntry(h, he->tag, he->t, he->p.ptr, he->c);
		he->append = 0;
                break;
            default:
                value = SvPV_nolen(ST(i));
		he->p.ptr = value;
		he->c = 1;
		he->append = 1;
                RETVAL = headerAddOrAppendEntry(h, he->tag, he->t, he->p.ptr, he->c); 
		he->append = 0;
                break;
        }
    }
    /* if (he->tag == RPMTAG_OLDFILENAMES) {
        compressFilelist(h); 
    } */
    OUTPUT:
    RETVAL
    
void
listtag(h)
    Header h
    PREINIT:
    HeaderIterator iterator;
    int tag;
    PPCODE:
    iterator = headerInitIterator(h);
    while (headerNextIterator(iterator, &tag, NULL, NULL, NULL)) {
        XPUSHs(sv_2mortal(newSViv(tag)));
    }
    headerFreeIterator(iterator);
    
int
hastag(h, sv_tag)
    Header h
    SV * sv_tag
    PREINIT:
    rpmTag tag = -1;
    CODE:
    if (SvIOK(sv_tag)) {
        tag = SvIV(sv_tag);
    } else if (SvPOK(sv_tag)) {
        tag = tagValue(SvPV_nolen(sv_tag));
    }    
    if (tag > 0)
        RETVAL = headerIsEntry(h, tag);
    else
        RETVAL = -1;
    OUTPUT:
    RETVAL
 
# Return the tag value in headers
void
tag(h, sv_tag)
    Header h
    SV * sv_tag
    PREINIT:
    HE_t he = memset(alloca(sizeof(*he)), 0, sizeof(*he));
    PPCODE:
    if (SvIOK(sv_tag)) {
        he->tag = SvIV(sv_tag);
    } else if (SvPOK(sv_tag)) {
        he->tag = tagValue(SvPV_nolen(sv_tag));
    }
    if (he->tag > 0)
        if (headerGetEntry(h, he->tag, &he->t, &he->p, &he->c)) {
            switch(he->t) {
                case RPM_STRING_ARRAY_TYPE:
                    {
                        int i;

                        EXTEND(SP, he->c);
        
                        for (i = 0; i < he->c; i++) {
                            PUSHs(sv_2mortal(newSVpv(he->p.argv[i], 0)));
                        }
                    }
                break;
                case RPM_STRING_TYPE:
                    PUSHs(sv_2mortal(newSVpv(he->p.str, 0)));
                break;
                case RPM_CHAR_TYPE:
                case RPM_UINT8_TYPE:
                case RPM_UINT16_TYPE:
                case RPM_UINT32_TYPE:
                    {
                        int i;

                        EXTEND(SP, he->c);

                        for (i = 0; i < he->c; i++) {
                            PUSHs(sv_2mortal(newSViv(he->p.ui32p[i])));
                        }
                    }
                break;
                case RPM_BIN_TYPE:
                    PUSHs(sv_2mortal(newSVpv(he->p.ptr, he->c)));
                break;
                default:
                    croak("unknown rpm tag type %d", he->t);
            }
        }
    he->p.ptr = headerFreeData(he->p.ptr, he->t);

int
tagtype(h, sv_tag)
    Header h
    SV * sv_tag
    PREINIT:
    HE_t he = memset(alloca(sizeof(*he)), 0, sizeof(*he));
    CODE:
    if (SvIOK(sv_tag)) {
        he->tag = SvIV(sv_tag);
    } else if (SvPOK(sv_tag)) {
        he->tag = tagValue(SvPV_nolen(sv_tag));
    }
    RETVAL = RPM_NULL_TYPE;
    if (he->tag > 0)
        if (headerGetEntry(h, he->tag, &he->t, NULL, &he->c))
            RETVAL = he->t;
    OUTPUT:
    RETVAL
    
rpmds
dependencies(header, sv_tag)
    Header header
    SV * sv_tag
    PREINIT:
    rpmTag tag;
    CODE:
    tag = sv2constant(sv_tag, "rpmtag");
    RETVAL = rpmdsNew(header, tag, 0);
    if (RETVAL) RETVAL = rpmdsInit(RETVAL);
    OUTPUT:
    RETVAL

rpmfi
files(header, ts = NULL)
    Header header
    rpmts ts
    PREINIT:
    rpmfi fi;
    CODE:
    if (ts)
        ts = rpmtsLink(ts, "RPM::Header::files");
    else
        ts = rpmtsCreate();
    RETVAL = rpmfiNew(ts, header, RPMTAG_BASENAMES, 0);
    if (RETVAL != NULL) RETVAL = rpmfiInit(RETVAL, 0);
    ts = rpmtsFree(ts);
    OUTPUT:
    RETVAL
 
